from typing import Optional

from elasticai.creator.hdl.code_generation.abstract_base_template import (
    module_to_package,
)
from elasticai.creator.hdl.design_base.design import Design, Port
from elasticai.creator.hdl.design_base.ports import (
    create_port_for_buffered_design as create_port,
)
from elasticai.creator.hdl.translatable import Path
from elasticai.creator.hdl.vhdl.code_generation.template import Template


class FPLinear1d(Design):
    @property
    def port(self) -> Port:
        return self._port

    def __init__(
        self,
        *,
        in_feature_num: int,
        out_feature_num: int,
        total_bits: int,
        frac_bits: int,
        work_library_name: str = "work",
        resource_option: str = "auto",
        name: Optional[str] = None,
    ):
        super().__init__(
            name="fp_linear1d" if name is None else name,
        )
        self._port = create_port(
            x_width=total_bits,
            y_width=total_bits,
            x_count=in_feature_num,
            y_count=out_feature_num,
        )
        self.in_feature_num = in_feature_num
        self.out_feature_num = out_feature_num
        self.work_library_name = work_library_name
        self.resource_option = resource_option
        self.frac_width = frac_bits
        self.data_width = total_bits
        self.x_addr_width = self.port["x_address"].width
        self.y_addr_width = self.port["y_address"].width

    def _template_parameters(self) -> dict[str, str]:
        return dict(
            (key, str(getattr(self, key)))
            for key in (
                "data_width",
                "frac_width",
                "x_addr_width",
                "y_addr_width",
                "in_feature_num",
                "out_feature_num",
            )
        )

    def save_to(self, destination: Path):
        template = Template(
            base_name="fp_linear_1d", package=module_to_package(self.__module__)
        )
        template.update_parameters(
            layer_name=self.name,
            work_library_name=self.work_library_name,
            resource_option=f'"{self.resource_option}"',
            **self._template_parameters(),
        )
        destination.as_file(f".vhd").write_text(template.lines())



from abc import abstractmethod
from typing import Protocol

from elasticai.creator.hdl.design_base.design import Design
from elasticai.creator.hdl.vhdl.designs.monotonously_increasing_precomputed_scalar_function.hard_sigmoid import (
    HardSigmoid,
)
from elasticai.creator.hdl.vhdl.number_representations import FixedPoint
from elasticai.creator.nn.hard_sigmoid import HardSigmoid as _HardSigmoidLayer


class FixedPointOps(Protocol):
    @property
    @abstractmethod
    def total_bits(self) -> int:
        ...

    @property
    @abstractmethod
    def frac_bits(self) -> int:
        ...


class FPHardSigmoid(_HardSigmoidLayer):
    def __init__(self, ops: FixedPointOps):
        super().__init__()
        self.ops = ops

    def translate(self) -> Design:
        def fp(value: float) -> int:
            return int(
                FixedPoint(
                    value=value,
                    total_bits=self.ops.total_bits,
                    frac_bits=self.ops.frac_bits,
                )
            )

        return HardSigmoid(
            lower_bound_for_zero=fp(-3),
            upper_bound_for_one=fp(3),
            width=self.ops.total_bits,
        )



from abc import abstractmethod
from typing import Protocol

from elasticai.creator.hdl.design_base.design import Design
from elasticai.creator.translatable_modules.module import Module as _BaseModule


class Tensor(Protocol):
    ...


class Module(_BaseModule[Design], Protocol):
    @abstractmethod
    def translate(self) -> Design:
        ...



from elasticai.creator.hdl.design_base.design import Design
from elasticai.creator.hdl.vhdl.designs.fp_linear_1d import FPLinear1d as FPLinearDesign
from elasticai.creator.nn.linear import FixedPointLinear


class FPLinear1d(FixedPointLinear):
    def translate(self) -> Design:
        return FPLinearDesign(
            frac_bits=self.frac_bits,
            total_bits=self.total_bits,
            in_feature_num=self.in_features,
            out_feature_num=self.out_features,
        )



from dataclasses import dataclass, field
from typing import Callable

from elasticai.creator.hdl.vhdl.code_files import (
    calculate_address_width,
    derive_fixed_point_params_from_factory,
)
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.templates import VHDLTemplate


@dataclass
class LSTMComponent:
    input_size: int
    hidden_size: int
    fixed_point_factory: Callable[[float], FixedPoint]
    layer_id: str
    work_library_name: str = field(default="work")

    def __post_init__(self) -> None:
        self.data_width, self.frac_width = derive_fixed_point_params_from_factory(
            self.fixed_point_factory
        )
        self.x_h_addr_width = calculate_address_width(
            self.input_size + self.hidden_size
        )
        self.hidden_addr_width = calculate_address_width(self.hidden_size)
        self.w_addr_width = calculate_address_width(
            (self.input_size + self.hidden_size) * self.hidden_size
        )

    @property
    def name(self) -> str:
        return f"lstm_{self.layer_id}.vhd"

    def lines(self) -> list[str]:
        template = VHDLTemplate(base_name="lstm")
        template.update_parameters(
            work_library_name=self.work_library_name,
            data_width=str(self.data_width),
            frac_width=str(self.frac_width),
            input_size=str(self.input_size),
            hidden_size=str(self.hidden_size),
            x_h_addr_width=str(self.x_h_addr_width),
            hidden_addr_width=str(self.hidden_addr_width),
            w_addr_width=str(self.w_addr_width),
            layer_name=self.layer_id,
        )
        return template.lines()



from collections import OrderedDict
from typing import cast

import torch.nn

from elasticai.creator.hdl.design_base.design import Design
from elasticai.creator.hdl.vhdl.designs.fp_linear_1d import (
    FPLinear1d as _FPLinear1dDesign,
)
from elasticai.creator.nn.lstm import FixedPointLSTMWithHardActivations as _nnLSTM
from elasticai.creator.translatable_modules.vhdl.fp_linear_1d import FPLinear1d
from elasticai.creator.translatable_modules.vhdl.lstm.design import LSTMNetworkDesign
from elasticai.creator.translatable_modules.vhdl.lstm.fp_lstm_cell import FPLSTMCell
from elasticai.creator.translatable_modules.vhdl.module import Module


class LSTMNetwork(torch.nn.Module):
    def __init__(self, layers: list[torch.nn.Module]):
        super().__init__()
        self.lstm = layers[0]
        self.layers = torch.nn.Sequential(
            OrderedDict(
                {f"fp_linear_1d_{i}": layer for i, layer in enumerate(layers[1:])}
            )
        )

    def translate(self) -> LSTMNetworkDesign:
        children = list(self.layers.children())
        first_lstm = cast(FixedPointLSTMWithHardActivations, self.lstm)
        total_bits = first_lstm.fixed_point_config.total_bits
        frac_bits = first_lstm.fixed_point_config.frac_bits
        hidden_size = first_lstm.hidden_size
        input_size = first_lstm.input_size
        follow_up_linear_layers = cast(
            list[_FPLinear1dDesign],
            [cast(FPLinear1d, layer).translate() for layer in self.layers],
        )
        return LSTMNetworkDesign(
            lstm=first_lstm.translate(),
            linear_layers=follow_up_linear_layers,
            total_bits=total_bits,
            frac_bits=frac_bits,
            hidden_size=hidden_size,
            input_size=input_size,
        )

    def forward(self, x):
        _, (x, _) = self.lstm(x)
        return self.layers(x)


class FixedPointLSTMWithHardActivations(_nnLSTM, Module):
    def translate(self) -> Design:
        def to_list(tensor: torch.Tensor) -> list:
            return tensor.detach().numpy().tolist()

        return FPLSTMCell(
            name="lstm_cell",
            total_bits=self.fixed_point_config.total_bits,
            frac_bits=self.fixed_point_config.frac_bits,
            w_ih=to_list(self.cell.linear_ih.weight),
            w_hh=to_list(self.cell.linear_hh.weight),
            b_ih=to_list(self.cell.linear_ih.bias),
            b_hh=to_list(self.cell.linear_hh.bias),
        )

    def __init__(
        self, total_bits: int, frac_bits: int, input_size: int, hidden_size: int
    ):
        super().__init__(
            total_bits=total_bits,
            frac_bits=frac_bits,
            input_size=input_size,
            hidden_size=hidden_size,
            bias=True,
            batch_first=True,
        )



import math
from abc import abstractmethod
from enum import Enum
from itertools import chain, filterfalse
from typing import (
    Callable,
    Iterable,
    Iterator,
    Literal,
    Optional,
    Protocol,
    Sequence,
    overload,
    runtime_checkable,
)

import torch.nn

from elasticai.creator.hdl.vhdl.code_generation.code_generation import (
    bin_representation,
)
from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    float_values_to_fixed_point,
    infer_total_and_frac_bits,
)

Code = Iterable[str]


@runtime_checkable
class CodeGenerator(Protocol):
    @abstractmethod
    def code(self) -> Iterable[str]:
        ...


CodeGeneratorCompatible = Code | CodeGenerator | str | Callable[[], Code]


def _vhdl_add_assignment(
    code: list[str], line_id: str, value: str, comment: Optional[str] = None
) -> None:
    new_code_fragment = f"{line_id} <= {bin_representation(value)};"
    if comment is not None:
        new_code_fragment += f" -- {comment}"
    code.append(new_code_fragment)


def _get_lower_case_class_name_or_component_name(
    cls: type, component_name: Optional[str]
) -> str:
    if component_name is None:
        return cls.__name__.lower()
    return component_name


def precomputed_scalar_function_process(
    x: list[FixedPoint], y: list[FixedPoint]
) -> list[str]:
    """
        returns the string of a lookup table
    Args:
        x (FixedPoint): input list
        y (FixedPoint) : output list
    Returns:
        String of lookup table (if/elsif statements for vhdl file)
    """
    x.sort()
    lines: list[str] = []
    if len(x) == 0 and len(y) == 1:
        _vhdl_add_assignment(
            code=lines,
            line_id="y",
            value=y[0].to_bin(),
        )
    elif len(x) != len(y) - 1:
        raise ValueError(
            f"x has to be one element shorter than y, but x has {len(x)} elements and y"
            f" {len(y)} elements"
        )
    else:
        smallest_possible_output = y[0]
        biggest_possible_output = y[-1]

        # first element
        for x_value in x[:1]:
            lines.append(f"if int_x<{x_value.to_signed_int()} then")
            _vhdl_add_assignment(
                code=lines,
                line_id="y",
                value=smallest_possible_output.to_bin(),
                comment=str(smallest_possible_output.to_signed_int()),
            )
            lines[-1] = "\t" + lines[-1]
        for current_x, current_y in zip(x[1:], y[1:-1]):
            lines.append(f"elsif int_x<{current_x.to_signed_int()} then")
            _vhdl_add_assignment(
                code=lines,
                line_id="y",
                value=current_y.to_bin(),
                comment=str(current_y.to_signed_int()),
            )
            lines[-1] = "\t" + lines[-1]
        # last element only in y
        for _ in y[-1:]:
            lines.append("else")
            _vhdl_add_assignment(
                code=lines,
                line_id="y",
                value=biggest_possible_output.to_bin(),
                comment=str(biggest_possible_output.to_signed_int()),
            )
            lines[-1] = "\t" + lines[-1]
        if len(lines) != 0:
            lines.append("end if;")

    # build the string block

    return lines


class PrecomputedScalarFunction:
    def __init__(
        self,
        x: list[FixedPoint],
        y: list[FixedPoint],
        component_name: Optional[str] = None,
    ):
        """
        We calculate the function with an algorithm equivalent to:
        ```
        def function(x: int, inputs: list[int], outputs: list[int]) -> int:
          for input, output in zip(inputs, outputs[:-1]):
            if x < input:
              return output
          return outputs[-1]
        ```
        """
        self.component_name = _get_lower_case_class_name_or_component_name(
            cls=type(self), component_name=component_name
        )

        self.data_width, self.frac_width = infer_total_and_frac_bits(x, y)
        self.x = x
        self.y = y
        self.process_instance = None

    @property
    def file_name(self) -> str:
        return f"{self.component_name}.vhd"

    def code(self) -> list[str]:
        library = ContextClause(
            library_clause=LibraryClause(logical_name_list=["ieee"]),
            use_clause=UseClause(
                selected_names=[
                    "ieee.std_logic_1164.all",
                    "ieee.numeric_std.all",
                ]
            ),
        )
        entity = Entity(self.component_name)
        entity.generic_list = [
            f"DATA_WIDTH : integer := {self.data_width}",
            f"FRAC_WIDTH : integer := {self.frac_width}",
        ]
        entity.port_list = [
            "x : in signed(DATA_WIDTH-1 downto 0)",
            "y : out signed(DATA_WIDTH-1 downto 0)",
        ]
        process = Process(
            identifier=self.component_name,
            lookup_table=precomputed_scalar_function_process(x=self.x, y=self.y),
            input_name="x",
        )
        process.process_declaration_list = ["variable int_x: integer := 0"]
        process.process_statements_list = ["int_x := to_integer(x)"]
        architecture = Architecture(
            design_unit=self.component_name,
        )
        architecture.architecture_statement_part = process
        code = list(chain(chain(library.code(), entity.code()), architecture.code()))
        return code


class Sigmoid(PrecomputedScalarFunction):
    def __init__(
        self, x: list[FixedPoint], component_name: Optional[str] = None
    ) -> None:
        x_tensor = torch.as_tensor(list(map(float, x)))
        # calculate y always for the previous element therefore, the last input is not needed here
        y = torch.nn.Sigmoid()(x_tensor[:-1]).tolist()
        y.insert(0, 0)
        # add last y value, therefore, x_tensor is one element shorter than y_tensor
        y.append(1)
        y = float_values_to_fixed_point(y, *infer_total_and_frac_bits(x))

        super(Sigmoid, self).__init__(x=x, y=y, component_name=component_name)


class Tanh(PrecomputedScalarFunction):
    def __init__(
        self, x: list[FixedPoint], component_name: Optional[str] = None
    ) -> None:
        y_list = [-1.0]
        # calculate y always for the previous element, therefore, the last input is not needed here
        for x_element in x[:-1]:
            y_list.append(math.tanh(float(x_element)))
        # add last y value, therefore, x_list is one element shorter than y_list
        y_list.append(1)
        y = float_values_to_fixed_point(y_list, *infer_total_and_frac_bits(x))

        super(Tanh, self).__init__(x=x, y=y, component_name=component_name)


Identifier = str


class Keywords(Enum):
    IS = "is"
    END = "end"
    ENTITY = "entity"
    COMPONENT = "component"
    PORT = "port"
    GENERIC = "generic"
    IN = "in"
    OUT = "out"
    INOUT = "inout"
    BUFFER = "buffer"
    LINKAGE = "linkage"
    INTEGER = "integer"
    STD_LOGIC = "std_logic"
    SIGNAL = "signal"
    ARCHITECTURE = "architecture"
    OF = "of"
    SIGNED = "signed"
    BEGIN = "begin"
    PROCESS = "process"
    STD_LOGIC_VECTOR = "std_logic_vector"


class DataType(Enum):
    INTEGER = Keywords.INTEGER.value
    STD_LOGIC = Keywords.STD_LOGIC.value
    SIGNED = Keywords.SIGNED.value
    STD_LOGIC_VECTOR = Keywords.STD_LOGIC_VECTOR.value


class Mode(Enum):
    IN = Keywords.IN.value
    OUT = Keywords.OUT.value
    INOUT = Keywords.INOUT.value
    BUFFER = Keywords.BUFFER.value


class _DesignUnitForEntityAndComponent:
    def __init__(
        self, identifier: str, design_type: Literal[Keywords.ENTITY, Keywords.COMPONENT]
    ):
        self.identifier = identifier
        self._generic_list = InterfaceList()
        self._port_list = InterfaceList()
        self.type = design_type

    @property
    def generic_list(self):
        return self._generic_list

    @generic_list.setter
    def generic_list(self, value):
        self._generic_list = InterfaceList(value)

    @property
    def port_list(self):
        return self._port_list

    @port_list.setter
    def port_list(self, value):
        self._port_list = InterfaceList(value)

    def _header(self) -> Iterator[str]:
        if len(self.generic_list) > 0:
            yield from _clause(Keywords.GENERIC, self._generic_list.code())
        if len(self.port_list) > 0:
            yield from _clause(Keywords.PORT, self._port_list.code())

    def __iter__(self) -> Iterator[str]:
        yield f"{self.type.value} {self.identifier} {Keywords.IS.value}"
        yield from _filter_empty_lines(self._header())
        yield f"{Keywords.END.value} {self.type.value} {self.identifier};"

    def code(self) -> Iterable[str]:
        return self


class Entity(_DesignUnitForEntityAndComponent):
    def __init__(self, identifier: str):
        super().__init__(identifier, Keywords.ENTITY)


class Architecture:
    def __init__(self, design_unit: str):
        self.design_unit = design_unit
        self.architecture_declaration_list = InterfaceList()
        self.architecture_component_list = InterfaceList()
        self.architecture_assignment_list = InterfaceList()
        self.architecture_process_list = InterfaceList()
        self.architecture_port_map_list = InterfaceList()
        self.architecture_assignment_at_end_of_declaration_list = InterfaceList()
        self.architecture_statement_part: Optional[CodeGenerator] = None

    def _generator(self) -> Iterator[str]:
        yield (
            f"{Keywords.ARCHITECTURE.value} rtl"
            f" {Keywords.OF.value} {self.design_unit} {Keywords.IS.value}"
        )
        if len(self.architecture_declaration_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(
                    self.architecture_declaration_list.code(), semicolon_last=True
                )
            )
        if len(self.architecture_component_list) > 0:
            yield from _filter_empty_lines(self.architecture_component_list.code())
        yield f"{Keywords.BEGIN.value}"
        if len(self.architecture_assignment_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(
                    self.architecture_assignment_list.code(), semicolon_last=True
                )
            )
        if len(self.architecture_process_list) > 0:
            yield from _filter_empty_lines(self.architecture_process_list.code())
        if len(self.architecture_port_map_list) > 0:
            yield from _filter_empty_lines(self.architecture_port_map_list.code())
        if len(self.architecture_assignment_at_end_of_declaration_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(
                    self.architecture_assignment_at_end_of_declaration_list.code(),
                    semicolon_last=True,
                )
            )

        if self.architecture_statement_part:
            yield from _filter_empty_lines(self.architecture_statement_part.code())
        yield f"{Keywords.END.value} {Keywords.ARCHITECTURE.value} rtl;"

    def code(self) -> Code:
        return self._generator()


class Process(Code, CodeGenerator):
    def __init__(
        self,
        identifier: str,
        input_name: Optional[str] = None,
        lookup_table: Optional[Code] = None,
    ):
        self.identifier = identifier
        self.process_declaration_list: list[str] = []
        self.process_statements_list: list[str] = []
        self.lookup_table = lookup_table
        self.input = input_name

    def _header(self) -> Iterator[str]:
        if len(self.process_declaration_list) > 0:
            yield from _append_semicolons_to_lines(self.process_declaration_list)

    def _body(self) -> Iterator[str]:
        if len(self.process_statements_list) > 0:
            yield from _append_semicolons_to_lines(self.process_statements_list)
        if self.lookup_table:
            yield from self.lookup_table

    def __iter__(self) -> Iterator[str]:
        if self.input:
            yield (
                f"{self.identifier}_{Keywords.PROCESS.value}:"
                f" {Keywords.PROCESS.value}({self.input})"
            )
        else:
            yield (
                f"{self.identifier}_{Keywords.PROCESS.value}: {Keywords.PROCESS.value}"
            )
        yield from _filter_empty_lines(self._header())
        yield f"{Keywords.BEGIN.value}"
        yield from _filter_empty_lines(self._body())
        yield (
            f"{Keywords.END.value} {Keywords.PROCESS.value} {self.identifier}_{Keywords.PROCESS.value};"
        )

    def code(self):
        return self


class ContextClause(Code, CodeGenerator):
    def __init__(self, library_clause, use_clause):
        self._use_clause = use_clause
        self._library_clause = library_clause

    def __iter__(self) -> Iterator[str]:
        yield from self._library_clause
        yield from self._use_clause

    def code(self) -> Code:
        return self


class UseClause(Code, CodeGenerator):
    def __init__(self, selected_names: list[str]):
        self._selected_names = selected_names

    def __iter__(self):
        def prefix_use(line: str):
            return f"use {line}"

        yield from _append_semicolons_to_lines(map(prefix_use, self._selected_names))

    def code(self) -> Code:
        return self


class LibraryClause(Code, CodeGenerator):
    def __init__(self, logical_name_list: list[str]):
        self._logical_name_list = logical_name_list

    def __iter__(self):
        yield from _append_semicolons_to_lines(
            ["library {}".format(", ".join(self._logical_name_list))]
        )

    def code(self) -> Code:
        return self


def _wrap_iterator_function_into_iterable(create_iterator: Callable[[], Iterator[str]]):
    class Wrapper(Iterable[str]):
        def __iter__(self) -> Iterator[str]:
            yield from create_iterator()

    wrapped = Wrapper()
    return wrapped


class CodeGeneratorConcatenation(Sequence[CodeGenerator], CodeGenerator):
    def __init__(self, *interfaces: CodeGeneratorCompatible):
        self.interface_generators: list[CodeGenerator] = [
            _unify_code_generators(interface) for interface in interfaces
        ]

    @overload
    def __getitem__(self, index: int) -> CodeGenerator:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[CodeGenerator]:
        ...

    def __getitem__(self, index):
        if isinstance(index, slice):
            return CodeGeneratorConcatenation(*tuple(self.interface_generators[index]))
        else:
            return self.interface_generators[index]

    def __len__(self) -> int:
        return len(self.interface_generators)

    def append(self, interface: CodeGeneratorCompatible) -> None:
        self.interface_generators.append(_unify_code_generators(interface))

    def _code_iterator(self) -> Iterator[str]:
        for generator in self.interface_generators:
            for line in generator.code():
                yield line

    def code(self) -> Code:
        return _wrap_iterator_function_into_iterable(self._code_iterator)


class InterfaceList(CodeGeneratorConcatenation):
    pass


ClauseType = Literal[Keywords.GENERIC, Keywords.COMPONENT, Keywords.PORT]


def _add_semicolons(lines: Code, semicolon_last: bool = False) -> Iterator[str]:
    temp = tuple(lines)
    yield from (f"{line};" for line in temp[:-1])
    yield f"{temp[-1]};" if semicolon_last else f"{temp[-1]}"


def _add_is(lines: Code) -> Iterator[str]:
    yield from (f"{line} is" for line in lines)


def _add_comma(lines: Code, comma_last: bool = False) -> Iterator[str]:
    temp = tuple(lines)
    yield from (f"{line}," for line in temp[:-1])
    yield f"{temp[-1]}," if comma_last else f"{temp[-1]}"


def _append_semicolons_to_lines(lines: Code) -> Iterator[str]:
    temp = tuple(lines)
    yield from (f"{line};" for line in temp)


def _clause(clause_type: ClauseType, interfaces: Code) -> Iterator[str]:
    yield f"{clause_type.value} ("
    yield from _filter_empty_lines(_add_semicolons(interfaces))
    yield ");"


def _filter_empty_lines(lines: Code) -> Iterator[str]:
    return filterfalse(_line_is_empty, lines)


def _line_is_empty(line: str) -> bool:
    return len(line) == 0


def _join_lines(lines) -> str:
    return "\n".join(lines)


def _empty_code_generator() -> Code:
    return []


def _wrap_in_is_end_block(
    block_type: Keywords, block_identifier: Identifier, lines: Code
) -> Iterator[str]:
    yield f"{block_type.value} {block_identifier} {Keywords.IS.value}"
    yield from _filter_empty_lines(lines)
    yield f"{Keywords.END.value} {block_type.value} {block_identifier};"


def _wrap_string_into_code_generator(string: str) -> CodeGenerator:
    class Wrapper(CodeGenerator):
        def code(self) -> Code:
            return [string]

    return Wrapper()


def _wrap_code_into_code_generator(code: Code) -> CodeGenerator:
    class Wrapped(CodeGenerator):
        def code(self) -> Code:
            return code

    wrapped = Wrapped()
    return wrapped


def _wrap_callable_into_code_generator(fn: Callable[[], Code]) -> CodeGenerator:
    class Wrapper(CodeGenerator):
        def code(self) -> Code:
            return fn()

    wrapped = Wrapper()
    return wrapped


def _unify_code_generators(generator: CodeGeneratorCompatible) -> CodeGenerator:
    if isinstance(generator, str):
        return _wrap_string_into_code_generator(generator)
    elif isinstance(generator, Iterable):
        return _wrap_code_into_code_generator(generator)
    elif isinstance(generator, CodeGenerator):
        return generator
    elif callable(generator):
        return _wrap_callable_into_code_generator(generator)
    else:
        raise ValueError



from elasticai.creator.vhdl.templates import VHDLTemplate


class DualPort2ClockRamComponent:
    def __init__(self, layer_id: str):
        self._name = "dual_port_2_clock_ram"
        self.template = VHDLTemplate(
            base_name=self._name,
        )
        self.id = layer_id

    @property
    def name(self) -> str:
        return f"{self._name}_{self.id}.vhd"

    def lines(self) -> list[str]:
        return self.template.lines()



from elasticai.creator.vhdl.number_representations import FixedPoint, FixedPointConfig
from elasticai.creator.vhdl.templates import VHDLTemplate


class FPHardTanhComponent:
    def __init__(
        self,
        min_val: FixedPoint,
        max_val: FixedPoint,
        fixed_point_factory: FixedPointConfig,
        layer_id: str,
    ):
        d = dict(
            data_width=fixed_point_factory.total_bits,
            frac_width=fixed_point_factory.frac_bits,
            min_val=min_val.to_signed_int(),
            max_val=max_val.to_signed_int(),
            layer_name=layer_id,
        )
        stringified_d = dict(((k, str(v)) for k, v in d.items()))
        self._name = f"fp_hard_tanh_{layer_id}"
        self.template = VHDLTemplate(base_name="fp_hard_tanh", **stringified_d)

    @property
    def single_line_parameters(self):
        return self.template.single_line_parameters

    @property
    def name(self) -> str:
        return f"{self._name}.vhd"

    def lines(self) -> list[str]:
        return self.template.lines()



from copy import copy
from functools import partial
from itertools import repeat
from typing import Any, cast

import numpy as np

from elasticai.creator.hdl.code_generation.abstract_base_template import (
    TemplateConfig,
    TemplateExpander,
    module_to_package,
)
from elasticai.creator.hdl.code_generation.code_generation import (
    calculate_address_width,
    to_hex,
)
from elasticai.creator.hdl.design_base.design import Design, Port
from elasticai.creator.hdl.design_base.signal import Signal
from elasticai.creator.hdl.translatable import Path
from elasticai.creator.hdl.vhdl.code_generation.code_generation import (
    generate_hex_for_rom,
)
from elasticai.creator.hdl.vhdl.designs import HardSigmoid
from elasticai.creator.nn._two_complement_fixed_point_config import (
    TwoComplementFixedPointConfig,
)
from elasticai.creator.translatable_modules.vhdl.lstm.fp_hard_tanh import FPHardTanh


class FPLSTMCell(Design):
    def __init__(
        self,
        *,
        name: str,
        total_bits: int,
        frac_bits: int,
        w_ih: list[list[list[float]]],
        w_hh: list[list[list[float]]],
        b_ih: list[list[float]],
        b_hh: list[list[float]],
    ):
        super().__init__(
            name=name,
        )
        base_config = TemplateConfig(
            package=module_to_package(self.__module__), file_name="", parameters={}
        )
        self.input_size = len(w_ih[0])
        self.hidden_size = len(w_ih) // 4
        self.weights_ih = w_ih
        self.weights_hh = w_hh
        self.biases_ih = b_ih
        self.biases_hh = b_hh
        self._fp_config = TwoComplementFixedPointConfig(
            total_bits=total_bits, frac_bits=frac_bits
        )
        self._rom_base_config = copy(base_config)
        self._rom_base_config.file_name = "rom.tpl.vhd"

        self._config = copy(base_config)
        self._config.file_name = f"{self.name}.tpl.vhd"
        self._config.parameters = {
            k: str(v)
            for k, v in dict(
                name=self.name,
                libary="work",
                data_width=total_bits,
                input_size=self.input_size,
                hidden_size=self.hidden_size,
                frac_width=frac_bits,
                x_h_addr_width=calculate_address_width(
                    self.input_size + self.hidden_size
                ),
                hidden_addr_width=calculate_address_width(self.hidden_size),
                w_addr_width=calculate_address_width(
                    (self.input_size + self.hidden_size) * self.hidden_size
                ),
            ).items()
        }

    @property
    def total_bits(self) -> int:
        return int(cast(str, self._config.parameters["data_width"]))

    @property
    def _hidden_addr_width(self) -> int:
        return int(cast(str, self._config.parameters["hidden_addr_width"]))

    @property
    def _weight_address_width(self) -> int:
        return int(cast(str, self._config.parameters["w_addr_width"]))

    @property
    def port(self) -> Port:
        ctrl_signal = partial(Signal, width=0)
        return Port(
            incoming=[
                Signal("x_data", self.total_bits),
                ctrl_signal("clock"),
                ctrl_signal("clk_hadamard"),
                ctrl_signal("reset"),
                ctrl_signal("zero_state"),
                ctrl_signal("h_out_en"),
                Signal("h_out_addr", self._hidden_addr_width),
            ],
            outgoing=[
                ctrl_signal("done"),
                Signal("h_out_data", self._hidden_addr_width),
            ],
        )

    def _build_weights(self):
        weights = np.concatenate((self.weights_ih, self.weights_hh), axis=1)
        weights = np.reshape(weights, (4, -1))
        w_i, w_f, w_g, w_o = weights.reshape(4, -1).tolist()

        bias = np.add(self.biases_ih, self.biases_hh)
        bias = np.reshape(bias, (4, -1))
        b_i, b_f, b_g, b_o = bias.tolist()

        def convert_floats_to_ints(floats: list[float]) -> list[int]:
            return list(map(self._fp_config.as_integer, floats))

        final_weights = tuple(map(convert_floats_to_ints, (w_i, w_f, w_g, w_o)))
        final_biases = tuple(map(convert_floats_to_ints, (b_i, b_f, b_g, b_o)))

        return final_weights, final_biases

    def save_to(self, destination: "Path"):
        weights, biases = self._build_weights()

        rom_template = TemplateExpander(self._rom_base_config)
        write_files = partial(
            self._write_files, destination=destination, rom_template=rom_template
        )
        write_files(
            names=("wi", "wf", "wg", "wo"),
            parameters=weights,
            address_width=self._weight_address_width,
        )
        write_files(
            names=("bi", "bf", "bg", "bo"),
            parameters=biases,
            address_width=self._hidden_addr_width,
        )
        self._save_dual_port_double_clock_ram(destination)
        self._save_hardtanh(destination)
        self._save_sigmoid(destination)
        expander = TemplateExpander(self._config)
        destination.create_subpath("lstm_cell").as_file(".vhd").write_text(
            expander.lines()
        )

    def _save_sigmoid(self, destination: Path):
        sigmoid_destination = destination.create_subpath("hard_sigmoid")
        sigmoid = HardSigmoid(
            width=self.total_bits,
            lower_bound_for_zero=self._fp_config.as_integer(-3),
            upper_bound_for_one=self._fp_config.as_integer(3),
        )
        sigmoid.save_to(sigmoid_destination)

    def _save_hardtanh(self, destination: Path):
        hardtanh_destination = destination.create_subpath("hard_tanh")
        hardtanh = FPHardTanh(
            total_bits=self.total_bits, frac_bits=self._fp_config.frac_bits
        )
        hardtanh.save_to(hardtanh_destination)

    def _write_files(
        self,
        destination: Path,
        rom_template: TemplateExpander,
        names: tuple[str, ...],
        parameters: Any,
        address_width: int,
    ):
        for values, name in zip(parameters, names):
            self._update_rom_base_config(
                name=name, values=values, address_width=address_width
            )
            rom_file = destination.create_subpath(f"{name}_rom").as_file(".vhd")
            rom_file.write_text(rom_template.lines())

    def _update_rom_base_config(self, values: list[int], name: str, address_width: int):
        values = self._pad_with_zeros(values, address_width)
        self._rom_base_config.parameters.update(
            dict(
                resource_option="auto",
                name=f"rom_{name}_{self.name}",
                rom_addr_bitwidth=str(address_width),
                rom_data_bitwidth=str(self.total_bits),
                rom_value=",".join(map(self._to_hex, values)),
            )
        )

    @staticmethod
    def _to_hex(value):
        return generate_hex_for_rom(to_hex(value, bit_width=8))

    @staticmethod
    def _pad_with_zeros(values, address_width):
        suffix = list(repeat(0, 2**address_width - len(values)))
        return values + suffix

    def _save_dual_port_double_clock_ram(self, destination: Path):
        template_configuration = TemplateConfig(
            file_name="dual_port_2_clock_ram.tpl.vhd",
            package=module_to_package(self.__module__),
            parameters=dict(
                name=f"{self.name}_dual_port_2_clock_ram",
            ),
        )
        template_expansion = TemplateExpander(template_configuration)

        destination.create_subpath(f"{self.name}_dual_port_2_clock_ram").as_file(
            ".vhd"
        ).write_text(template_expansion.lines())



from typing import cast

from torch.nn import Module as _torchModule
from torch.nn import Sequential as torchSequential

from elasticai.creator.hdl.design_base.design import Design
from elasticai.creator.hdl.vhdl.designs.sequential import (
    Sequential as _SequentialDesign,
)

from .module import Module


class Sequential(torchSequential):
    def __init__(self, submodules: tuple[Module]):
        super().__init__(*cast(tuple[_torchModule], submodules))

    def translate(self) -> Design:
        submodules: list[Module] = [cast(Module, m) for m in self.children()]
        subdesigns = [m.translate() for m in submodules]
        x_address_width = 1
        y_address_width = 1
        if len(subdesigns) == 0:
            x_width = 1
            y_width = 1

        else:
            front = subdesigns[0]
            back = subdesigns[-1]
            x_width = front.port["x"].width
            y_width = back.port["y"].width
            found_y_address = False
            found_x_address = False
            for design in subdesigns:
                if "y_address" in design.port and not found_y_address:
                    found_y_address = True
                    y_address_width = design.port["y_address"].width
                if "x_address" in design.port and not found_x_address:
                    found_x_address = True
                    x_address_width = back.port["x_address"].width
        return _SequentialDesign(
            sub_designs=subdesigns,
            x_width=x_width,
            y_width=y_width,
            x_address_width=x_address_width,
            y_address_width=y_address_width,
        )



from abc import ABC, abstractmethod
from functools import partial
from itertools import chain

from elasticai.creator.hdl.code_generation.abstract_base_template import (
    module_to_package,
)
from elasticai.creator.hdl.code_generation.code_generation import (
    calculate_address_width,
)
from elasticai.creator.hdl.design_base import std_signals
from elasticai.creator.hdl.design_base.design import Design, Port
from elasticai.creator.hdl.design_base.signal import Signal
from elasticai.creator.hdl.design_base.std_signals import (
    clock,
    done,
    enable,
    x,
    x_address,
    y,
    y_address,
)
from elasticai.creator.hdl.translatable import Path
from elasticai.creator.hdl.vhdl.code_generation import create_instance
from elasticai.creator.hdl.vhdl.code_generation.code_generation import (
    create_connections,
    create_signal_definitions,
)
from elasticai.creator.hdl.vhdl.code_generation.template import Template


class Sequential(Design):
    def __init__(
        self,
        sub_designs: list[Design],
        *,
        x_width: int,
        y_width: int,
        x_address_width: int,
        y_address_width: int,
    ):
        super().__init__("Sequential")
        self._x_width = x_width
        self._y_width = y_width
        self._y_address_width = y_address_width
        self._x_address_width = x_address_width
        self._instances: dict[str, Design] = {}
        self._names: dict[str, int] = {}
        self._library_name_for_instances = "work"
        self._architecture_name_for_instances = "rtl"
        self._autowirer = _AutoWirer
        for design in sub_designs:
            self._register_subdesign(design)

    def _make_name_unique(self, name: str) -> str:
        return f"i_{name}_{self._get_counter_for_name(name)}"

    def _get_counter_for_name(self, name: str) -> int:
        if name in self._names:
            return self._names[name]
        else:
            return 0

    def _increment_name_counter(self, name: str):
        self._names[name] = 1 + self._get_counter_for_name(name)

    def _register_subdesign(self, d: Design):
        unique_name = self._make_name_unique(d.name)
        self._instances[unique_name] = d
        self._increment_name_counter(d.name)

    def _qualified_signal_name(self, instance: str, signal: str) -> str:
        return f"{instance}_{signal}"

    @property
    def port(self) -> Port:
        return Port(
            incoming=[
                x(width=self._x_width),
                y_address(width=self._y_address_width),
                clock(),
                enable(),
            ],
            outgoing=[
                y(width=self._y_width),
                x_address(width=self._x_address_width),
                done(),
            ],
        )

    def _save_subdesigns(self, destination: Path) -> None:
        for name, design in self._instances.items():
            design.save_to(destination.create_subpath(name))

    def _create_dataflow_nodes(self) -> list["_DataFlowNode"]:
        nodes: list[_DataFlowNode] = [
            _StartNode(x_width=self._x_width, y_address_width=self._y_address_width)
        ]
        for instance, design in self._instances.items():
            nodes.append(self._create_data_flow_node(instance, design))
        nodes.append(
            _EndNode(y_width=self._y_width, x_address_width=self._x_address_width)
        )
        return nodes

    @staticmethod
    def _create_data_flow_node(instance: str, design: Design) -> "_DataFlowNode":
        node = _InstanceNode(instance=instance)
        node.add_sinks(design.port.incoming)
        node.add_sources(design.port.outgoing)
        return node

    def _generate_connections(self) -> list[str]:
        nodes = self._create_dataflow_nodes()
        wirer = self._autowirer(nodes=nodes)
        return create_connections(wirer.connect())

    def _generate_instantiations(self) -> list[str]:
        instantiations: list[str] = list()
        for instance, design in self._instances.items():
            signal_map = {
                signal.name: self._qualified_signal_name(instance, signal.name)
                for signal in design.port
            }
            instantiations.extend(
                create_instance(
                    name=instance,
                    entity=design.name,
                    library=self._library_name_for_instances,
                    architecture=self._architecture_name_for_instances,
                    signal_mapping=signal_map,
                )
            )
        return instantiations

    def _generate_signal_definitions(self) -> list[str]:
        return sorted(
            chain.from_iterable(
                create_signal_definitions(f"{instance_id}_", instance.port.signals)
                for instance_id, instance in self._instances.items()
            )
        )

    def save_to(self, destination: Path):
        network_implementation = Template(
            "network", package=module_to_package(self.__module__)
        )
        self._save_subdesigns(destination)
        network_implementation.update_parameters(
            layer_connections=self._generate_connections(),
            layer_instantiations=self._generate_instantiations(),
            signal_definitions=self._generate_signal_definitions(),
            x_address_width=str(self._x_address_width),
            y_address_width=str(self._y_address_width),
            x_width=str(self._x_width),
            y_width=str(self._y_width),
        )
        target_file = destination.create_subpath("sequential").as_file(".vhd")
        target_file.write_text(network_implementation.lines())


class _AutoWirer:
    def __init__(self, nodes: list["_DataFlowNode"]):
        self.nodes = nodes
        self.mapping = {
            "x": ["x", "y"],
            "y": ["y", "x"],
            "x_address": ["x_address", "y_address"],
            "y_address": ["y_address", "x_address"],
            "enable": ["enable", "done"],
            "done": ["done", "enable"],
            "clock": ["clock"],
        }
        self.available_sources: dict[str, "_OwnedSignal"] = {}

    def _pick_best_matching_source(self, sink: "_OwnedSignal") -> "_OwnedSignal":
        for source_name in self.mapping[sink.name]:
            if source_name in self.available_sources:
                source = self.available_sources[source_name]
                return source
        return _TopNode().sources[0]

    def _update_available_sources(self, node: "_DataFlowNode") -> None:
        self.available_sources.update({s.name: s for s in node.sources})

    def connect(self) -> dict[str, str]:
        connections: dict[str, str] = {}
        for node in self.nodes:
            for sink in node.sinks:
                source = self._pick_best_matching_source(sink)
                connections[sink.qualified_name] = source.qualified_name
            self._update_available_sources(node)

        return connections


class _DataFlowNode(ABC):
    def __init__(self) -> None:
        self.sinks: list[_OwnedSignal] = []
        self.sources: list[_OwnedSignal] = []

    def add_sinks(self, sinks: list[Signal]):
        create_sink = partial(_OwnedSignal, owner=self)
        self.sinks.extend(map(create_sink, sinks))

    def add_sources(self, sources: list[Signal]):
        create_source = partial(_OwnedSignal, owner=self)
        self.sources.extend(map(create_source, sources))

    @property
    @abstractmethod
    def prefix(self) -> str:
        ...


class _TopNode(_DataFlowNode):
    @property
    def prefix(self) -> str:
        return ""


class _StartNode(_TopNode):
    def __init__(self, x_width: int, y_address_width: int):
        super().__init__()
        self.add_sources(
            [
                std_signals.x(x_width),
                std_signals.y_address(y_address_width),
                std_signals.enable(),
                std_signals.clock(),
            ]
        )

    @property
    def prefix(self) -> str:
        return ""


class _EndNode(_TopNode):
    def __init__(self, y_width: int, x_address_width: int):
        super().__init__()
        self.add_sinks(
            [
                std_signals.y(y_width),
                std_signals.x_address(x_address_width),
                std_signals.done(),
            ]
        )


class _InstanceNode(_DataFlowNode):
    def __init__(self, instance: str):
        self.instance = instance
        super().__init__()

    @property
    def prefix(self) -> str:
        return f"{self.instance}_"


class _OwnedSignal:
    def __init__(self, signal: Signal, owner: _DataFlowNode):
        self.owner = owner
        self._signal = signal

    @property
    def name(self) -> str:
        return self._signal.name

    @property
    def width(self) -> int:
        return self._signal.width

    @property
    def qualified_name(self) -> str:
        return f"{self.owner.prefix}{self.name}"



from elasticai.creator.hdl.code_generation.abstract_base_template import (
    TemplateConfig,
    TemplateExpander,
)


class Template:
    def __init__(
        self,
        base_name: str,
        package: str = "elasticai.creator.hdl.vhdl.template_resources",
        suffix: str = ".tpl.vhd",
    ):
        self._template_name = base_name
        self._internal_template = TemplateExpander(
            TemplateConfig(
                file_name=f"{base_name}{suffix}", package=package, parameters=dict()
            )
        )

    def update_parameters(self, **parameters: str | list[str]):
        self._internal_template.config.parameters.update(parameters)

    def lines(self) -> list[str]:
        return self._internal_template.lines()



from typing import Callable

from elasticai.creator.vhdl.number_representations import FixedPoint


def derive_fixed_point_params_from_factory(
    fixed_point_factory: Callable[[float], FixedPoint]
) -> tuple[int, int]:
    dummy_value = fixed_point_factory(0)
    return dummy_value.total_bits, dummy_value.frac_bits



from dataclasses import dataclass

from elasticai.creator.vhdl.number_representations import FixedPointConfig, parameters
from elasticai.creator.vhdl.templates import VHDLTemplate


@dataclass
class FPReLUComponent:
    layer_id: str  # used to distinguish layers in the same model
    fixed_point_factory: FixedPointConfig

    def __post_init__(self) -> None:
        self.data_width, self.frac_width = parameters(self.fixed_point_factory)

    @property
    def name(self) -> str:
        return f"fp_relu_{self.layer_id}.vhd"

    def lines(self) -> list[str]:
        template = VHDLTemplate(base_name="fp_relu")
        template.update_parameters(
            layer_name=self.layer_id,
            data_width=str(self.data_width),
            clock_option="false",
        )

        return template.lines()



import warnings
from dataclasses import dataclass

from elasticai.creator.vhdl.code.code_file import CodeFile
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.precomputed_scalar_function import Tanh


@dataclass
class TanhComponent(CodeFile):
    x: list[FixedPoint]

    @property
    def name(self) -> str:
        return f"tanh.vhd"

    def lines(self) -> list[str]:
        return list(Tanh(x=self.x, component_name="tanh").code())

    def code(self) -> list[str]:
        warnings.warn(
            message=DeprecationWarning(
                (
                    f"calling instance directly is deprecated, use the"
                    f" lines() method instead "
                ),
            ),
            stacklevel=2,
        )
        return self.lines()



from dataclasses import dataclass, field
from typing import Callable

from elasticai.creator.hdl.vhdl.code_files import (
    calculate_address_width,
    derive_fixed_point_params_from_factory,
)
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.templates import VHDLTemplate


@dataclass
class FPLinear1dFile:
    def save_to(self, prefix: str):
        raise NotImplementedError()

    layer_id: str  # used to distinguish layers in the same model
    in_feature_num: int
    out_feature_num: int
    fixed_point_factory: Callable[[float], FixedPoint]
    work_library_name: str = field(default="work")
    resource_option: str = "auto"

    def __post_init__(self) -> None:
        self.data_width, self.frac_width = derive_fixed_point_params_from_factory(
            self.fixed_point_factory
        )
        self.x_addr_width = calculate_address_width(self.in_feature_num)
        self.y_addr_width = calculate_address_width(self.out_feature_num)

    def _template_parameters(self) -> dict[str, str]:
        return dict(
            (key, str(getattr(self, key)))
            for key in (
                "data_width",
                "frac_width",
                "x_addr_width",
                "y_addr_width",
                "in_feature_num",
                "out_feature_num",
            )
        )

    @property
    def name(self) -> str:
        return f"fp_linear_1d_{self.layer_id}.vhd"

    def code(self) -> list[str]:
        template = VHDLTemplate(base_name="fp_linear_1d")
        template.update_parameters(
            layer_name=self.layer_id,
            work_library_name=self.work_library_name,
            resource_option=f'"{self.resource_option}"',
            **self._template_parameters(),
        )
        return template.lines()



from functools import partial

from elasticai.creator.hdl.code_generation.abstract_base_template import (
    TemplateConfig,
    TemplateExpander,
    module_to_package,
)
from elasticai.creator.hdl.code_generation.code_generation import (
    calculate_address_width,
)
from elasticai.creator.hdl.design_base import std_signals
from elasticai.creator.hdl.design_base.design import Design, Port
from elasticai.creator.hdl.design_base.signal import Signal
from elasticai.creator.hdl.translatable import Path
from elasticai.creator.hdl.vhdl.designs.fp_linear_1d import FPLinear1d


class LSTMNetworkDesign(Design):
    def save_to(self, destination: "Path"):
        self._lstm.save_to(destination.create_subpath("lstm_cell"))
        for index, layer in enumerate(self._linear_layers):
            layer.save_to(destination.create_subpath(f"fp_linear_1d_{index}"))
        expander = TemplateExpander(self.config)
        destination.create_subpath("lstm_network").as_file(".vhd").write_text(
            expander.lines()
        )

    def __init__(
        self,
        lstm: Design,
        linear_layers: list[FPLinear1d],
        total_bits: int,
        frac_bits: int,
        hidden_size: int,
        input_size: int,
    ):
        super().__init__(name="lstm_network")
        self._linear_layers = linear_layers
        signal = partial(Signal)
        self._lstm = lstm
        ctrl_signal = partial(Signal, width=0)
        self.config = TemplateConfig(
            module_to_package(self.__module__),
            file_name="lstm_network.tpl.vhd",
            parameters=dict(
                data_width=str(total_bits),
                frac_width=str(frac_bits),
                hidden_size=str(hidden_size),
                input_size=str(input_size),
                linear_in_features=str(self._linear_layers[0].in_feature_num),
                linear_out_features=str(self._linear_layers[0].out_feature_num),
                hidden_addr_width=(
                    f"{calculate_address_width(hidden_size + input_size)}"
                ),
                x_h_addr_width=f"{calculate_address_width(hidden_size + input_size)}",
                w_addr_width=(
                    f"{calculate_address_width((hidden_size + input_size) * hidden_size)}"
                ),
                in_addr_width="4",
            ),
        )
        self._port = Port(
            incoming=[
                std_signals.clock(),
                std_signals.enable(),
                ctrl_signal("x_we"),
                signal("x_in", lstm.port["x_data"].width),
                signal("addr_in", lstm.port["h_out_addr"].width),
            ],
            outgoing=[
                ctrl_signal("done"),
                signal("d_out", lstm.port["h_out_data"].width),
            ],
        )

    @property
    def port(self) -> Port:
        return self._port


class LSTMNetworkSkeleton(Design):
    def __init__(self, data_width, data_address_width):
        super().__init__("lstm_network")
        self._port = self._create_port(data_width, data_address_width)

    def _create_port(self, data_width: int, data_address_width: int) -> Port:
        _signal = Signal
        ctr_signal = partial(Signal, width=0)
        incoming_control_signals = [
            ctr_signal(name=name)
            for name in ("clock", "clk_hadamard", "reset", "rd", "wr")
        ]
        incoming_data_signals = [
            _signal(name="data_in", width=data_width),
            _signal(name="address_in", width=data_address_width),
        ]
        outgoing_data_signals = [
            _signal(name="data_out", width=data_width),
            _signal(name="debug", width=8),
            _signal(name="led_ctrl", width=4),
        ]
        outgoing_control_signals = [
            ctr_signal(name="busy"),
            ctr_signal(name="wake_up"),
        ]
        return Port(
            incoming=incoming_data_signals + incoming_control_signals,
            outgoing=outgoing_control_signals + outgoing_data_signals,
        )

    @property
    def port(self) -> Port:
        return self._port

    def save_to(self, destination: "Path"):
        pass



import warnings
from dataclasses import dataclass

from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.precomputed_scalar_function import Sigmoid


@dataclass
class SigmoidComponent:
    x: list[FixedPoint]

    @property
    def name(self) -> str:
        return f"sigmoid.vhd"

    def lines(self) -> list[str]:
        return list(Sigmoid(x=self.x, component_name="sigmoid").code())

    def code(self) -> list[str]:
        warnings.warn(
            message=DeprecationWarning(
                f"call is deprecated, use the lines() method instead ",
            ),
            stacklevel=2,
        )
        return self.lines()



import math
from collections.abc import Iterator, Sequence
from itertools import chain
from typing import Any, Callable


def _assert_range(value: float, total_bits: int, frac_bits: int) -> None:
    max_value = 2 ** (total_bits - frac_bits - 1)
    min_value = max_value * (-1)
    if not min_value <= value < max_value:
        raise ValueError(
            f"Value {value} cannot represented as a fixed point value with"
            f" {total_bits} total bits and {frac_bits} fraction bits (value range:"
            f" [{min_value}, {max_value}))."
        )


def _assert_is_compatible(fp1: "FixedPoint", fp2: "FixedPoint") -> None:
    if not (fp1.total_bits == fp2.total_bits and fp1.frac_bits == fp2.frac_bits):
        raise ValueError(
            f"FixedPoint objects not compatible (total_bits: {fp1.total_bits} !="
            f" {fp2.total_bits}); frac_bits: {fp1.frac_bits} != {fp2.frac_bits})."
        )


def _invert_int(value: int, num_bits: int) -> int:
    return value ^ int("1" * num_bits, 2)


def _discard_leading_bits(value: int, num_bits: int) -> int:
    return value & int("1" * num_bits, 2)


def _calculate_two_complement(value: int, num_bits: int) -> int:
    return _invert_int(abs(value), num_bits) + 1


class FixedPoint:
    """
    A data type that converts a given number to the corresponding fixed-point representation.
    A fixed-point value is an unsigned integer in two's complement.
    Parameters:
        value (float): Value to be represented as fixed-point value.
        total_bits (int): Total number of bits of the fixed-point representation (including number of fractional bits).
        frac_bits (int): Number of bits to represent the fractional part of the number.
    Examples:
        >>> fixed_point_value = FixedPoint(-2.9, total_bits=8, frac_bits=4)
        >>> int(fixed_point_value)
        210
        >>> float(fixed_point_value)
        -2.875
        >>> fixed_point_value.to_signed_int()
        -46
        >>> fixed_point_value.to_bin()
        '11010010'
        >>> fixed_point_value.to_hex()
        'd2'
    """

    __slots__ = ["_value", "_frac_bits", "_total_bits"]

    def __init__(self, value: float, total_bits: int, frac_bits: int) -> None:
        self._value = float(value)
        self._total_bits = total_bits
        self._frac_bits = frac_bits
        _assert_range(self._value, self._total_bits, self._frac_bits)

    def __int__(self) -> int:
        fp_int = int(self._value * (1 << self._frac_bits))
        if fp_int < 0:
            fp_int = _calculate_two_complement(fp_int, self._total_bits)
        return fp_int

    def __float__(self) -> float:
        return FixedPoint.from_unsigned_int(
            int(self), self._total_bits, self._frac_bits
        )._value

    def __eq__(self, other: Any) -> bool:
        return float(self) == float(other)

    def __ne__(self, other: Any) -> bool:
        return float(self) != float(other)

    def __lt__(self, other: Any) -> bool:
        return float(self) < float(other)

    def __le__(self, other: Any) -> bool:
        return float(self) <= float(other)

    def __gt__(self, other: Any) -> bool:
        return float(self) > float(other)

    def __ge__(self, other: Any) -> bool:
        return float(self) >= float(other)

    def __add__(self, other: "FixedPoint") -> "FixedPoint":
        _assert_is_compatible(self, other)
        return self._identical_fixed_point_from_int(
            _discard_leading_bits(int(self) + int(other), num_bits=self._total_bits)
        )

    def __sub__(self, other: "FixedPoint") -> "FixedPoint":
        return self + (-other)

    def __and__(self, other: "FixedPoint") -> "FixedPoint":
        _assert_is_compatible(self, other)
        return self._identical_fixed_point_from_int(int(self) & int(other))

    def __or__(self, other: "FixedPoint") -> "FixedPoint":
        _assert_is_compatible(self, other)
        return self._identical_fixed_point_from_int(int(self) | int(other))

    def __xor__(self, other: "FixedPoint") -> "FixedPoint":
        _assert_is_compatible(self, other)
        return self._identical_fixed_point_from_int(int(self) ^ int(other))

    def __invert__(self) -> "FixedPoint":
        return self._identical_fixed_point_from_int(
            _invert_int(int(self), num_bits=self._total_bits)
        )

    def __neg__(self) -> "FixedPoint":
        return self._identical_fixed_point(-self._value)

    def __abs__(self) -> "FixedPoint":
        return self._identical_fixed_point(abs(self._value))

    def __str__(self) -> str:
        return str(int(self))

    def __repr__(self) -> str:
        return (
            f"FixedPoint(value={self._value}, total_bits={self._total_bits},"
            f" frac_bits={self._frac_bits})"
        )

    def _identical_fixed_point(self, value: float) -> "FixedPoint":
        return FixedPoint(
            value=value, total_bits=self._total_bits, frac_bits=self._frac_bits
        )

    def _identical_fixed_point_from_int(self, value: int) -> "FixedPoint":
        return FixedPoint.from_unsigned_int(
            value=value, total_bits=self._total_bits, frac_bits=self._frac_bits
        )

    @staticmethod
    def from_unsigned_int(value: int, total_bits: int, frac_bits: int) -> "FixedPoint":
        if value > 2**total_bits - 1:
            raise ValueError(
                f"Value {value} cannot interpreted as a fixed point with"
                f" {total_bits} total bits."
            )
        is_negative = value & (1 << total_bits - 1) > 0
        if is_negative:
            value = _calculate_two_complement(value, total_bits)
            value *= -1
        float_value = value / (1 << frac_bits)
        return FixedPoint(float_value, total_bits=total_bits, frac_bits=frac_bits)

    @staticmethod
    def from_signed_int(value: int, total_bits: int, frac_bits: int) -> "FixedPoint":
        float_value = value / (1 << frac_bits)
        return FixedPoint(float_value, total_bits=total_bits, frac_bits=frac_bits)

    @staticmethod
    def get_builder(total_bits: int, frac_bits: int) -> "FixedPointConfig":
        return FixedPointConfig(
            total_bits=total_bits, frac_bits=frac_bits, constructor=FixedPoint
        )

    @property
    def total_bits(self) -> int:
        return self._total_bits

    @property
    def frac_bits(self) -> int:
        return self._frac_bits

    def bin_iter(self) -> Iterator[int]:
        return ((int(self) >> i) & 1 for i in range(self._total_bits))

    def to_signed_int(self) -> int:
        return int(abs(self)) * (-1 if self < 0 else 1)

    def to_bin(self) -> str:
        return f"{int(self):0{self._total_bits}b}"

    def to_hex(self) -> str:
        return f"{int(self):0{math.ceil(self._total_bits / 4)}x}"


class ClippedFixedPoint(FixedPoint):
    def __init__(self, value: float, total_bits: int, frac_bits: int) -> None:
        max_value = (2 ** (total_bits - 1) - 1) / (1 << frac_bits)
        min_value = 2 ** (total_bits - frac_bits - 1) * (-1)
        if min_value <= value <= max_value:
            super().__init__(value=value, total_bits=total_bits, frac_bits=frac_bits)
        else:
            super().__init__(
                value=max_value if value > max_value else min_value,
                total_bits=total_bits,
                frac_bits=frac_bits,
            )

    def __float__(self) -> float:
        return ClippedFixedPoint.from_unsigned_int(
            int(self), self._total_bits, self._frac_bits
        )._value

    def __repr__(self) -> str:
        return (
            f"ClippedFixedPoint(value={self._value}, total_bits={self._total_bits},"
            f" frac_bits={self._frac_bits})"
        )

    def _identical_fixed_point(self, value: float) -> "ClippedFixedPoint":
        return ClippedFixedPoint(
            value=value, total_bits=self._total_bits, frac_bits=self._frac_bits
        )

    def _identical_fixed_point_from_int(self, value: int) -> "ClippedFixedPoint":
        return ClippedFixedPoint.from_unsigned_int(
            value=value, total_bits=self._total_bits, frac_bits=self._frac_bits
        )

    @staticmethod
    def from_unsigned_int(
        value: int, total_bits: int, frac_bits: int
    ) -> "ClippedFixedPoint":
        value = _discard_leading_bits(value, num_bits=total_bits)
        is_negative = value & (1 << total_bits - 1) > 0
        if is_negative:
            value = _calculate_two_complement(value, total_bits)
            value *= -1
        float_value = value / (1 << frac_bits)
        return ClippedFixedPoint(
            float_value, total_bits=total_bits, frac_bits=frac_bits
        )

    @staticmethod
    def from_signed_int(
        value: int, total_bits: int, frac_bits: int
    ) -> "ClippedFixedPoint":
        float_value = value / (1 << frac_bits)
        return ClippedFixedPoint(
            float_value, total_bits=total_bits, frac_bits=frac_bits
        )

    @staticmethod
    def get_builder(total_bits: int, frac_bits: int) -> "FixedPointConfig":
        return FixedPointConfig(
            constructor=ClippedFixedPoint, total_bits=total_bits, frac_bits=frac_bits
        )


def infer_total_and_frac_bits(*values: Sequence[FixedPoint]) -> tuple[int, int]:
    if sum(len(value_list) == 0 for value_list in values) > 0:
        raise ValueError("Cannot infer total bits and frac bits from an empty list.")
    total_bits, frac_bits = values[0][0].total_bits, values[0][0].frac_bits
    for value in chain(*values):
        if value.total_bits != total_bits or value.frac_bits != frac_bits:
            raise ValueError(
                "Cannot infer total bits and frac bits from a list with mixed total"
                " bits or frac bits."
            )
    return total_bits, frac_bits


class FixedPointConfig:
    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        constructor: Callable[[float, int, int], FixedPoint],
    ):
        self.frac_bits = frac_bits
        self.total_bits = total_bits
        self._constructor = constructor

    def __call__(self, f: float) -> FixedPoint:
        return self._constructor(f, self.total_bits, self.frac_bits)


def parameters(factory: FixedPointConfig) -> tuple[int, int]:
    return factory.total_bits, factory.frac_bits


def float_values_to_fixed_point(
    values: list[float], total_bits: int, frac_bits: int
) -> list[FixedPoint]:
    return list(map(lambda x: FixedPoint(x, total_bits, frac_bits), values))


def unsigned_int_values_to_fixed_point(
    values: list[int], total_bits: int, frac_bits: int
) -> list[FixedPoint]:
    return list(
        map(lambda x: FixedPoint.from_unsigned_int(x, total_bits, frac_bits), values)
    )



from abc import abstractmethod
from typing import Protocol, TypeVar

T_contra = TypeVar("T_contra", contravariant=True)


class Acceptor(Protocol[T_contra]):
    @abstractmethod
    def accepts(self, source: T_contra) -> bool:
        ...



from dataclasses import dataclass

from elasticai.creator.vhdl.number_representations import FixedPoint, FixedPointConfig
from elasticai.creator.vhdl.templates import VHDLTemplate


@dataclass
class FPHardSigmoidComponent:
    layer_id: str
    zero_threshold: FixedPoint
    one_threshold: FixedPoint
    slope: FixedPoint
    y_intercept: FixedPoint
    fixed_point_factory: FixedPointConfig

    @property
    def name(self) -> str:
        return f"fp_hard_sigmoid_{self.layer_id}.vhd"

    def lines(self) -> list[str]:
        code = VHDLTemplate(
            base_name="fp_hard_sigmoid",
            data_width=str(self.fixed_point_factory.total_bits),
            frac_width=str(self.fixed_point_factory.frac_bits),
            one=str(self.fixed_point_factory(1).to_signed_int()),
            zero_threshold=str(self.zero_threshold.to_signed_int()),
            one_threshold=str(self.one_threshold.to_signed_int()),
            y_intercept=str(self.y_intercept.to_signed_int()),
            slope=str(self.slope.to_signed_int()),
            layer_name=self.layer_id,
        )

        return code.lines()



class CodeFileBase:
    @property
    def name(self) -> str:
        return self._name

    def lines(self) -> list[str]:
        return self._code

    def __repr__(self) -> str:
        return f"CodeBaseFile(name={self._name}, code={self._code})"

    def __init__(self, name: str, code: list[str]):
        self._name = name
        self._code = code
